@@ -1,21 +1,21 @@
 /*
- * This file is part of gspell.
+ * This file is part of gspell, a spell-checking library.
  *
  * Copyright 2002 - Paolo Maggi
  * Copyright 2015 - Sébastien Wilmet
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
- * This program is distributed in the hope that it will be useful,
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, see <http://www.gnu.org/licenses/>.
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this library; if not, see <http://www.gnu.org/licenses/>.
  */
 
 /* This is a modified version of GtkSpell 2.0.5 (gtkspell.sf.net)
@@ -57,7 +57,8 @@ struct _GspellInlineCheckerGtv
 	/* List of GtkTextView* */
 	GSList *views;
 
-	GtkTextTag *tag_highlight;
+	GtkTextTag *highlight_tag;
+	GtkTextTag *no_spell_check_tag;
 
 	GtkTextMark *mark_click;
 
@@ -114,14 +115,14 @@ check_word (GspellInlineCheckerGtv *spell,
 
 	if (error != NULL)
 	{
-		g_warning ("Automatic spell checker: %s", error->message);
+		g_warning ("Inline spell checker: %s", error->message);
 		g_error_free (error);
 	}
 
 	if (!correctly_spelled)
 	{
 		gtk_text_buffer_apply_tag (spell->buffer,
-					   spell->tag_highlight,
+					   spell->highlight_tag,
 					   start,
 					   end);
 	}
@@ -130,77 +131,68 @@ check_word (GspellInlineCheckerGtv *spell,
 }
 
 static void
-adjust_iters_at_word_boundaries (GtkTextIter *start,
-				 GtkTextIter *end)
+check_subregion (GspellInlineCheckerGtv *spell,
+		 GtkTextIter            *start,
+		 GtkTextIter            *end)
 {
-	if (gtk_text_iter_inside_word (end))
+	GtkTextIter word_start;
+
+	g_assert_cmpint (gtk_text_iter_compare (start, end), <=, 0);
+
+	if (gtk_text_iter_inside_word (start) &&
+	    !gtk_text_iter_starts_word (start))
+	{
+		gtk_text_iter_backward_word_start (start);
+	}
+
+	if (gtk_text_iter_inside_word (end) &&
+	    !gtk_text_iter_starts_word (end))
 	{
 		gtk_text_iter_forward_word_end (end);
 	}
 
-	if (!gtk_text_iter_starts_word (start))
-	{
-		if (gtk_text_iter_inside_word (start) ||
-		    gtk_text_iter_ends_word (start))
-		{
-			gtk_text_iter_backward_word_start (start);
-		}
-		else
-		{
-			/* If we're neither at the beginning nor inside a word,
-			 * me must be in some spaces.
-			 * Skip forward to the beginning of the next word.
-			 */
-			if (gtk_text_iter_forward_word_end (start) ||
-			    gtk_text_iter_is_end (start))
-			{
-				gtk_text_iter_backward_word_start (start);
-			}
-		}
-	}
-}
-
-static void
-check_subregion (GspellInlineCheckerGtv *spell,
-		 const GtkTextIter      *start,
-		 const GtkTextIter      *end)
-{
-	GtkTextIter start_adjusted;
-	GtkTextIter end_adjusted;
-	GtkTextIter word_start;
-
-	start_adjusted = *start;
-	end_adjusted = *end;
-	adjust_iters_at_word_boundaries (&start_adjusted, &end_adjusted);
-
 	gtk_text_buffer_remove_tag (spell->buffer,
-				    spell->tag_highlight,
-				    &start_adjusted,
-				    &end_adjusted);
+				    spell->highlight_tag,
+				    start,
+				    end);
 
-	word_start = start_adjusted;
+	word_start = *start;
+	if (!gtk_text_iter_starts_word (&word_start))
+	{
+		gtk_text_iter_forward_word_end (&word_start);
 
-	while (_gspell_utils_skip_no_spell_check (&word_start, &end_adjusted) &&
-	       gtk_text_iter_compare (&word_start, &end_adjusted) < 0)
+		/* Didn't move, there is no words after @start_adjusted. */
+		if (gtk_text_iter_equal (&word_start, start))
+		{
+			return;
+		}
+
+		gtk_text_iter_backward_word_start (&word_start);
+		g_assert (gtk_text_iter_starts_word (&word_start));
+		g_assert_cmpint (gtk_text_iter_compare (start, &word_start), <, 0);
+	}
+
+	while (_gspell_utils_skip_no_spell_check (spell->no_spell_check_tag, &word_start, end) &&
+	       gtk_text_iter_compare (&word_start, end) < 0)
 	{
 		GtkTextIter word_end;
 		GtkTextIter next_word_start;
 
+		g_assert (gtk_text_iter_starts_word (&word_start));
+
 		word_end = word_start;
 		gtk_text_iter_forward_word_end (&word_end);
 
-		if (gtk_text_iter_starts_word (&word_start) &&
-		    gtk_text_iter_ends_word (&word_end))
-		{
-			check_word (spell, &word_start, &word_end);
-		}
+		g_assert_cmpint (gtk_text_iter_compare (&word_end, end), <=, 0);
+
+		check_word (spell, &word_start, &word_end);
 
 		next_word_start = word_end;
 		gtk_text_iter_forward_word_end (&next_word_start);
 		gtk_text_iter_backward_word_start (&next_word_start);
 
-		/* Make sure we've actually advanced (we don't advance if we are
-		 * at the end of the buffer).
+		/* Make sure we've actually advanced (we don't advance if we
+		 * have just checked the last word of the buffer).
 		 */
 		if (gtk_text_iter_compare (&next_word_start, &word_start) <= 0)
 		{
@@ -211,35 +203,6 @@ check_subregion (GspellInlineCheckerGtv *spell,
 	}
 }
 
-static void
-check_region (GspellInlineCheckerGtv *spell,
-	      GtkTextRegion          *region)
-{
-	GtkTextRegionIterator region_iter;
-
-	if (region == NULL)
-	{
-		return;
-	}
-
-	gtk_text_region_get_iterator (region, &region_iter, 0);
-
-	while (!gtk_text_region_iterator_is_end (&region_iter))
-	{
-		GtkTextIter start;
-		GtkTextIter end;
-
-		if (!gtk_text_region_iterator_get_subregion (&region_iter, &start, &end))
-		{
-			return;
-		}
-
-		check_subregion (spell, &start, &end);
-
-		gtk_text_region_iterator_next (&region_iter);
-	}
-}
-
 static void
 get_visible_region (GtkTextView *view,
 		    GtkTextIter *start,
@@ -311,6 +274,7 @@ check_visible_region_in_view (GspellInlineCheckerGtv *spell,
 	GtkTextIter visible_start;
 	GtkTextIter visible_end;
 	GtkTextRegion *intersect;
+	GtkTextRegionIterator intersect_iter;
 
 	if (spell->scan_region == NULL)
 	{
@@ -328,12 +292,26 @@ check_visible_region_in_view (GspellInlineCheckerGtv *spell,
 		return;
 	}
 
-	check_region (spell, intersect);
-	gtk_text_region_destroy (intersect);
+	gtk_text_region_get_iterator (intersect, &intersect_iter, 0);
 
-	gtk_text_region_subtract (spell->scan_region,
-				  &visible_start,
-				  &visible_end);
+	while (!gtk_text_region_iterator_is_end (&intersect_iter))
+	{
+		GtkTextIter start;
+		GtkTextIter end;
+
+		if (!gtk_text_region_iterator_get_subregion (&intersect_iter, &start, &end))
+		{
+			return;
+		}
+
+		check_subregion (spell, &start, &end);
+
+		gtk_text_region_subtract (spell->scan_region, &start, &end);
+
+		gtk_text_region_iterator_next (&intersect_iter);
+	}
+
+	gtk_text_region_destroy (intersect);
 
 	if (is_text_region_empty (spell->scan_region))
 	{
@@ -402,21 +380,12 @@ add_subregion_to_scan (GspellInlineCheckerGtv *spell,
 		       const GtkTextIter      *start,
 		       const GtkTextIter      *end)
 {
-	GtkTextIter start_adjusted;
-	GtkTextIter end_adjusted;
-
 	if (spell->scan_region == NULL)
 	{
 		spell->scan_region = gtk_text_region_new (spell->buffer);
 	}
 
-	start_adjusted = *start;
-	end_adjusted = *end;
-	adjust_iters_at_word_boundaries (&start_adjusted, &end_adjusted);
-
-	gtk_text_region_add (spell->scan_region,
-			     &start_adjusted,
-			     &end_adjusted);
+	gtk_text_region_add (spell->scan_region, start, end);
 }
 
 static void
@@ -454,7 +423,28 @@ delete_range_after_cb (GtkTextBuffer          *buffer,
 		       GtkTextIter            *end,
 		       GspellInlineCheckerGtv *spell)
 {
-	add_subregion_to_scan (spell, start, end);
+	GtkTextIter start_adjusted;
+	GtkTextIter end_adjusted;
+
+	start_adjusted = *start;
+	end_adjusted = *end;
+
+	/* Just to be sure. Normally start == end. */
+	gtk_text_iter_order (&start_adjusted, &end_adjusted);
+
+	if (gtk_text_iter_ends_word (&start_adjusted) ||
+	    (gtk_text_iter_inside_word (&start_adjusted) &&
+	     !gtk_text_iter_starts_word (&start_adjusted)))
+	{
+		gtk_text_iter_backward_word_start (&start_adjusted);
+	}
+
+	if (gtk_text_iter_inside_word (&end_adjusted))
+	{
+		gtk_text_iter_forward_word_end (&end_adjusted);
+	}
+
+	add_subregion_to_scan (spell, &start_adjusted, &end_adjusted);
 	install_timeout (spell, TIMEOUT_DURATION_BUFFER_MODIFIED);
 }
 
@@ -674,7 +664,7 @@ populate_popup_cb (GtkTextView            *view,
 		return;
 	}
 
-	if (!gtk_text_iter_has_tag (&start, spell->tag_highlight))
+	if (!gtk_text_iter_has_tag (&start, spell->highlight_tag))
 	{
 		return;
 	}
@@ -737,7 +727,7 @@ remove_tag_to_word (GspellInlineCheckerGtv *spell,
 		    gtk_text_iter_ends_word (&match_end))
 		{
 			gtk_text_buffer_remove_tag (spell->buffer,
-						    spell->tag_highlight,
+						    spell->highlight_tag,
 						    &match_start,
 						    &match_end);
 		}
@@ -747,16 +737,16 @@ remove_tag_to_word (GspellInlineCheckerGtv *spell,
 }
 
 static void
-add_word_cb (GspellChecker          *checker,
-	     const gchar            *word,
-	     GspellInlineCheckerGtv *spell)
+word_added_cb (GspellChecker          *checker,
+	       const gchar            *word,
+	       GspellInlineCheckerGtv *spell)
 {
 	remove_tag_to_word (spell, word);
 }
 
 static void
-clear_session_cb (GspellChecker          *checker,
-		  GspellInlineCheckerGtv *spell)
+session_cleared_cb (GspellChecker          *checker,
+		    GspellInlineCheckerGtv *spell)
 {
 	recheck_all (spell);
 }
@@ -817,13 +807,18 @@ popup_menu_cb (GtkTextView            *view,
 }
 
 static void
-update_tag_highlight_priority (GspellInlineCheckerGtv *spell,
-			       GtkTextTagTable        *table)
+apply_or_remove_tag_cb (GtkTextBuffer          *buffer,
+			GtkTextTag             *tag,
+			GtkTextIter            *start,
+			GtkTextIter            *end,
+			GspellInlineCheckerGtv *spell)
 {
-	g_return_if_fail (spell->tag_highlight != NULL);
-
-	gtk_text_tag_set_priority (spell->tag_highlight,
-				   gtk_text_tag_table_get_size (table) - 1);
+	if (spell->no_spell_check_tag != NULL &&
+	    spell->no_spell_check_tag == tag)
+	{
+		add_subregion_to_scan (spell, start, end);
+		install_timeout (spell, TIMEOUT_DURATION_BUFFER_MODIFIED);
+	}
 }
 
 static void
@@ -831,7 +826,20 @@ tag_added_cb (GtkTextTagTable        *table,
 	      GtkTextTag             *tag,
 	      GspellInlineCheckerGtv *spell)
 {
-	update_tag_highlight_priority (spell, table);
+	gchar *name;
+
+	g_object_get (tag, "name", &name, NULL);
+
+	if (g_strcmp0 (name, "gtksourceview:context-classes:no-spell-check") == 0)
+	{
+		g_return_if_fail (spell->no_spell_check_tag == NULL);
+
+		spell->no_spell_check_tag = g_object_ref (tag);
+
+		recheck_all (spell);
+	}
+
+	g_free (name);
 }
 
 static void
@@ -839,41 +847,26 @@ tag_removed_cb (GtkTextTagTable        *table,
 		GtkTextTag             *tag,
 		GspellInlineCheckerGtv *spell)
 {
-	if (tag != spell->tag_highlight)
+	if (spell->no_spell_check_tag != NULL &&
+	    spell->no_spell_check_tag == tag)
 	{
-		update_tag_highlight_priority (spell, table);
+		g_clear_object (&spell->no_spell_check_tag);
+
+		recheck_all (spell);
 	}
 }
 
-static void
-tag_changed_cb (GtkTextTagTable        *table,
-		GtkTextTag             *tag,
-		gboolean                size_changed,
-		GspellInlineCheckerGtv *spell)
-{
-	update_tag_highlight_priority (spell, table);
-}
-
-static void
-highlight_updated_cb (GtkSourceBuffer        *buffer,
-		      GtkTextIter            *start,
-		      GtkTextIter            *end,
-		      GspellInlineCheckerGtv *spell)
-{
-	add_subregion_to_scan (spell, start, end);
-	install_timeout (spell, TIMEOUT_DURATION_BUFFER_MODIFIED);
-}
-
 static void
 set_buffer (GspellInlineCheckerGtv *spell,
-	    GtkSourceBuffer        *buffer)
+	    GtkTextBuffer          *buffer)
 {
 	GtkTextTagTable *tag_table;
 	GtkTextIter start;
 
-	g_return_if_fail (GTK_SOURCE_IS_BUFFER (buffer));
+	g_return_if_fail (GTK_IS_TEXT_BUFFER (buffer));
 	g_return_if_fail (spell->buffer == NULL);
-	g_return_if_fail (spell->tag_highlight == NULL);
+	g_return_if_fail (spell->highlight_tag == NULL);
+	g_return_if_fail (spell->no_spell_check_tag == NULL);
 	g_return_if_fail (spell->mark_click == NULL);
 
 	spell->buffer = g_object_ref (buffer);
@@ -895,15 +888,27 @@ set_buffer (GspellInlineCheckerGtv *spell,
 				 G_CONNECT_AFTER);
 
 	g_signal_connect_object (buffer,
-				 "highlight-updated", /* GtkSourceBuffer signal */
-				 G_CALLBACK (highlight_updated_cb),
+				 "apply-tag",
+				 G_CALLBACK (apply_or_remove_tag_cb),
 				 spell,
-				 0);
+				 G_CONNECT_AFTER);
 
-	spell->tag_highlight = gtk_text_buffer_create_tag (spell->buffer, NULL,
+	g_signal_connect_object (buffer,
+				 "remove-tag",
+				 G_CALLBACK (apply_or_remove_tag_cb),
+				 spell,
+				 G_CONNECT_AFTER);
+
+	spell->highlight_tag = gtk_text_buffer_create_tag (spell->buffer, NULL,
 							   "underline", PANGO_UNDERLINE_ERROR,
 							   NULL);
-	g_object_ref (spell->tag_highlight);
+	g_object_ref (spell->highlight_tag);
+
+	spell->no_spell_check_tag = _gspell_utils_get_no_spell_check_tag (spell->buffer);
+	if (spell->no_spell_check_tag != NULL)
+	{
+		g_object_ref (spell->no_spell_check_tag);
+	}
 
 	tag_table = gtk_text_buffer_get_tag_table (spell->buffer);
 
@@ -919,12 +924,6 @@ set_buffer (GspellInlineCheckerGtv *spell,
 				 spell,
 				 0);
 
-	g_signal_connect_object (tag_table,
-				 "tag-changed",
-				 G_CALLBACK (tag_changed_cb),
-				 spell,
-				 0);
-
 	/* For now we don't care where the mark points. The start looks like a
 	 * good place to begin with.
 	 */
@@ -948,20 +947,20 @@ set_spell_checker (GspellInlineCheckerGtv *spell,
 	_gspell_checker_check_language_set (checker);
 
 	g_signal_connect_object (spell->spell_checker,
-				 "add_word_to_session",
-				 G_CALLBACK (add_word_cb),
+				 "word-added-to-session",
+				 G_CALLBACK (word_added_cb),
 				 spell,
 				 0);
 
 	g_signal_connect_object (spell->spell_checker,
-				 "add_word_to_personal",
-				 G_CALLBACK (add_word_cb),
+				 "word-added-to-personal",
+				 G_CALLBACK (word_added_cb),
 				 spell,
 				 0);
 
 	g_signal_connect_object (spell->spell_checker,
-				 "clear_session",
-				 G_CALLBACK (clear_session_cb),
+				 "session-cleared",
+				 G_CALLBACK (session_cleared_cb),
 				 spell,
 				 0);
 
@@ -1033,9 +1032,9 @@ gspell_inline_checker_gtv_dispose (GObject *object)
 
 		table = gtk_text_buffer_get_tag_table (spell->buffer);
 
-		if (table != NULL && spell->tag_highlight != NULL)
+		if (table != NULL && spell->highlight_tag != NULL)
 		{
-			gtk_text_tag_table_remove (table, spell->tag_highlight);
+			gtk_text_tag_table_remove (table, spell->highlight_tag);
 		}
 
 		if (spell->mark_click != NULL)
@@ -1050,7 +1049,8 @@ gspell_inline_checker_gtv_dispose (GObject *object)
 		spell->buffer = NULL;
 	}
 
-	g_clear_object (&spell->tag_highlight);
+	g_clear_object (&spell->highlight_tag);
+	g_clear_object (&spell->no_spell_check_tag);
 	g_clear_object (&spell->spell_checker);
 
 	g_slist_free_full (spell->views, g_object_unref);
@@ -1085,7 +1085,7 @@ gspell_inline_checker_gtv_class_init (GspellInlineCheckerGtvClass *klass)
 	/**
 	 * GspellInlineCheckerGtv:buffer:
 	 *
-	 * The #GtkSourceBuffer. If a same buffer is used for several views, the
+	 * The #GtkTextBuffer. If a same buffer is used for several views, the
 	 * misspelled words are visible in all views, because #GtkTextTag's are
 	 * added to the buffer.
 	 */
@@ -1094,7 +1094,7 @@ gspell_inline_checker_gtv_class_init (GspellInlineCheckerGtvClass *klass)
 					 g_param_spec_object ("buffer",
 							      "Buffer",
 							      "",
-							      GTK_SOURCE_TYPE_BUFFER,
+							      GTK_TYPE_TEXT_BUFFER,
 							      G_PARAM_READWRITE |
 							      G_PARAM_CONSTRUCT_ONLY |
 							      G_PARAM_STATIC_STRINGS));
@@ -1122,18 +1122,18 @@ gspell_inline_checker_gtv_init (GspellInlineCheckerGtv *spell)
 
 /**
  * gspell_inline_checker_gtv_new:
- * @buffer: a #GtkSourceBuffer.
+ * @buffer: a #GtkTextBuffer.
  * @checker: a #GspellChecker.
  *
  * Returns: a new #GspellInlineCheckerGtv object.
  */
 GspellInlineCheckerGtv *
-gspell_inline_checker_gtv_new (GtkSourceBuffer *buffer,
-			       GspellChecker   *checker)
+gspell_inline_checker_gtv_new (GtkTextBuffer *buffer,
+			       GspellChecker *checker)
 {
 	GspellInlineCheckerGtv *spell;
 
-	g_return_val_if_fail (GTK_SOURCE_IS_BUFFER (buffer), NULL);
+	g_return_val_if_fail (GTK_IS_TEXT_BUFFER (buffer), NULL);
 	g_return_val_if_fail (GSPELL_IS_CHECKER (checker), NULL);
 
 	spell = g_object_get_data (G_OBJECT (buffer), INLINE_CHECKER_GTV_KEY);
